#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
/* For 64bit kernel / 32bit userspace */
#include <linux/netfilter_ipv4/ipt_MNAT.h>

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"MNAT v%s options:\n"
" --to <template>\n"
"				Addresses to map source to.\n"
"				Every template has the form\n"
"	        <bytetmpl>.<bytetmpl>.<bytetmpl>.<bytetmpl>\n"
"				Every bytetmpl has the form\n"
"			                          /12%%34+56\n",
IPTABLES_VERSION);
}

static struct option opts[] = {
	{ "to", 1, 0, '1' },
	{ 0 }
};

/* Initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache)
{
	struct ipt_mnat_template *tmpl = (struct ipt_mnat_template *)t->data;
	int i;

	for(i = 0; i < 4; i++) {
		tmpl->byte_template[i].divisor = 1;
		tmpl->byte_template[i].base = 1;
		tmpl->byte_template[i].offset = 0;
	}

	/* Can't cache this */
	*nfcache |= NFC_UNKNOWN;
}

/* Parses range of IPs */
static void
parse_to(char *arg, struct ipt_mnat_byte_template *template)
{
	int count;
	char *a;
	unsigned int tmp;
	
	if((a = strchr(arg, '/'))) {
		count = sscanf(a, "/%d", &tmp);
		if(count < 1) {
			exit_error(PARAMETER_PROBLEM,
				   "divisor did not parse: %s", a);
		}
		template->divisor = tmp;
	}
	if((a = strchr(arg, '%'))) {
		count = sscanf(a, "%%%d", &tmp);
		if(count < 1) {
			exit_error(PARAMETER_PROBLEM,
				   "base did not parse: %s", a);
		}
		template->base = tmp;
	}
	if((a = strchr(arg, '+'))) {
		count = sscanf(a, "+%d", &tmp);
		if(count < 1) {
			exit_error(PARAMETER_PROBLEM,
				   "offset did not parse: %s", a);
		}
		template->offset = tmp;
	}
}

#define IPT_SAME_OPT_TO			0x01

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target)
{
	struct ipt_mnat_template *tmpl
		= (struct ipt_mnat_template *)(*target)->data;
	char *strtok_buf;
	char *nextbyte;
	int i;

	switch (c) {
	case '1':
		nextbyte = strtok_r(optarg, ",:;.", &strtok_buf);
		for(i = 0; i < 4; i++) {
			if(!nextbyte) {
				exit_error(PARAMETER_PROBLEM,
					   "invalid MNAT --to specification - need more templates");
			}
			parse_to(nextbyte, tmpl->byte_template + (3 - i));
			nextbyte = strtok_r(NULL, ",:;.", &strtok_buf);
		}

		*flags |= IPT_SAME_OPT_TO;
		break;
	default:
		return 0;
	}
	
	return 1;
}

/* Final check; need --to. */
static void final_check(unsigned int flags)
{
	if (!(flags & IPT_MNAT_OPT_TO))
		exit_error(PARAMETER_PROBLEM,
			   "MNAT needs --to");
}

/* Saves the union ipt_targinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_target *target)
{
	int pos;
	struct ipt_mnat_template *tmpl
		= (struct ipt_mnat_template *)target->data;
	printf("--to ");

	for (pos = 0; pos < 4; pos++) {
		printf("/%d%%%d+%d",
				tmpl->byte_template[pos].divisor,
				tmpl->byte_template[pos].base,
				tmpl->byte_template[pos].offset
				);
		if(pos < 3)
			printf(".");
	}
}

/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target,
      int numeric)
{
	save(ip, target);
}

static
struct iptables_target mnat
= { NULL,
    "MNAT",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_mnat_template)),
    IPT_ALIGN(sizeof(struct ipt_mnat_template)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_target(&mnat);
}
